import os
from skimage import io
import torch
from torch.autograd import Variable
import torch.nn.functional as F
from torch.utils.data import DataLoader
from torchvision import transforms

import numpy as np
from PIL import Image

from data_loaderd import RescaleT
from data_loaderd import ToTensorLab
from data_loaderd import SalObjDataset

from misc import crf_refine

from model_depth import GlassNet


import torch.backends.cudnn as cudnn

cudnn.enabled = True
cudnn.benchmark = True
torch.cuda.set_device(0)

import time

def normPRED(d):
    ma = torch.max(d)
    mi = torch.min(d)
    dn = (d - mi) / (ma - mi)
    return dn


def save_reflection(image_name, pred, d_dir):
    predict = pred
    predict = predict.squeeze()
    predict_np = predict.cpu().data.numpy()
    predict_np = np.clip(predict_np, 0, 1)
    predict_np = np.transpose(predict_np, (1, 2, 0)) * 255.
    im = Image.fromarray((predict_np.astype(np.uint8))).convert('RGB')

    img_name = image_name.split("/")[-1]
    image = io.imread(image_name)
    imo = im.resize((image.shape[1], image.shape[0]), resample=Image.BILINEAR)

    aaa = img_name.split(".")
    bbb = aaa[0:-1]
    imidx = bbb[0]
    for i in range(1, len(bbb)):
        imidx = imidx + "." + bbb[i]

    imo.save(d_dir + imidx + '.png')

def save_output(image_name, input, pred, d_dir, crf=True):
    predict = pred
    predict = predict.squeeze()
    predict_np = predict.cpu().data.numpy()

    if crf:
        input = input.cpu().data.numpy().squeeze() * 255
        input = np.transpose(input, (1, 2, 0)).astype(np.uint8)
        predict_np = (predict_np * 255).astype(np.uint8)

        predict_np = predict_np.copy(order='C')
        input = input.copy(order='C')

        predict_np = crf_refine(input, predict_np)

        im = Image.fromarray(predict_np).convert('RGB')
    else:
        im = Image.fromarray(predict_np * 255).convert('RGB')

    img_name = image_name.split("/")[-1]
    image = io.imread(image_name)
    imo = im.resize((image.shape[1], image.shape[0]), resample=Image.BILINEAR)

    aaa = img_name.split(".")
    bbb = aaa[0:-1]
    imidx = bbb[0]
    for i in range(1, len(bbb)):
        imidx = imidx + "." + bbb[i]

    imo.save(d_dir + imidx + '.png')


# --------- 1. get image path and name ---------

from sys import argv

# script, exp_name = argv

import importlib

exp_name = "model_depth"
ablation_name = "best_multiscale"



# exp_name = "depth_20220107"
image_dir = '../datasets/glass_depth/test/images/'
depth_dir = '../datasets/glass_depth/test/depths/'
prediction_root = '../our_ablation_results/{}/'.format(ablation_name)
reflection_root = '../our_ablation_results/{}_reflections/'.format(ablation_name)


img_name_list = [os.path.join(image_dir, img_name) for img_name in os.listdir(image_dir)]

depth_name_list = []
for img_path in img_name_list:
    img_name = img_path.split("/")[-1]

    aaa = img_name.split(".")
    bbb = aaa[0:-1]
    imidx = bbb[0]
    for i in range(1, len(bbb)):
        imidx = imidx + "." + bbb[i]

    depth_name_list.append(depth_dir + imidx + '.png')

# --------- 2. load data ---------
# 1. load data
test_salobj_dataset = SalObjDataset(img_name_list=img_name_list, lbl_name_list=[], dep_name_list=depth_name_list,
                                    transform=transforms.Compose([RescaleT(384), ToTensorLab(flag=0)]))
test_salobj_dataloader = DataLoader(test_salobj_dataset, batch_size=1, shuffle=False, num_workers=1)




# --------- 3. model define ---------
print("...load GlassNet...")
net = GlassNet()

if torch.cuda.is_available():
    net.cuda()

from glob import glob

# all_models = glob("./saved_models/{}/*".format(exp_name))
best_model = "best.pth"

model_name = os.path.basename(best_model)
# model_dir = './saved_models/glassnet_depth/' + model_name

prediction_dir = prediction_root + model_name.split('_')[3] + '/' + 'glass_depth/'
if not os.path.exists(prediction_dir):
    os.makedirs(prediction_dir)

reflection_dir = reflection_root + model_name.split('_')[3] + '/' + 'glass_depth/'
if not os.path.exists(reflection_dir):
    os.makedirs(reflection_dir)

net.load_state_dict(torch.load(best_model))

net.eval()

# --------- 4. inference for each image ---------
for i_test, data_test in enumerate(test_salobj_dataloader):

    print("inferring:", img_name_list[i_test].split("/")[-1])

    inputs_test = data_test['image']
    inputs_test = inputs_test.type(torch.FloatTensor)
    
    inputs_depth = inputs_test[:, 0, :, :].unsqueeze(1)
    # inputs_depth = torch.zeros((1,1,384,384))
    # inputs_depth = data_test['depth']
    # inputs_depth = inputs_depth.type(torch.FloatTensor)

    # inputs_dm = data_test['depthmissing']
    # inputs_dm = inputs_dm.type(torch.FloatTensor)


    inputs_dm = inputs_depth
    if torch.cuda.is_available():
        inputs_test = Variable(inputs_test.cuda())
        inputs_depth = Variable(inputs_depth.cuda())
        inputs_dm = Variable(inputs_dm.cuda())


    else:
        inputs_test = Variable(inputs_test)
        inputs_depth = Variable(inputs_depth)
        inputs_dm = Variable(inputs_dm)


        
    start = time.time()
    d1, d2, d3, d4, depth1, depth2, depth3, depth4 = net(inputs_test, inputs_depth, inputs_dm)
    end = time.time()
    print(end-start)
    # normalization
    pred = d1[:, 0, :, :]
    pred = F.sigmoid(pred)
    pred = normPRED(pred)
    
    # save results to test_results folder
    
    prediction_dir = prediction_root + model_name.split('_')[3] + '/' + 'glass_depth/' + 'd1/'
    if not os.path.exists(prediction_dir):
        os.makedirs(prediction_dir)
    
    save_output(img_name_list[i_test], inputs_test, pred, prediction_dir, False)
    
    pred = d2[:, 0, :, :]
    pred = F.sigmoid(pred)
    pred = normPRED(pred)
    
    
    prediction_dir = prediction_root + model_name.split('_')[3] + '/' + 'glass_depth/' + 'd2/'
    if not os.path.exists(prediction_dir):
        os.makedirs(prediction_dir)    
    # save results to test_results folder
    save_output(img_name_list[i_test], inputs_test, pred, prediction_dir, False)

    prediction_dir = prediction_root + model_name.split('_')[3] + '/' + 'glass_depth/' + 'd3/'
    if not os.path.exists(prediction_dir):
        os.makedirs(prediction_dir)    
    pred = d3[:, 0, :, :]
    pred = F.sigmoid(pred)
    pred = normPRED(pred)


    # save results to test_results folder
    save_output(img_name_list[i_test], inputs_test, pred, prediction_dir, False)
    
    pred = d4[:, 0, :, :]
    pred = F.sigmoid(pred)
    pred = normPRED(pred)

    prediction_dir = prediction_root + model_name.split('_')[3] + '/' + 'glass_depth/' + 'd4/'
    if not os.path.exists(prediction_dir):
        os.makedirs(prediction_dir)     
    # save results to test_results folder
    save_output(img_name_list[i_test], inputs_test, pred, prediction_dir, False)
    
    # save_reflection(img_name_list[i_test], ref, reflection_dir)

    del  d1, d2, d3, d4
