import torch
from torch.autograd import Variable
import torch.nn as nn

from torch.utils.data import DataLoader
from torchvision import transforms
import torch.optim as optim
import os

import glob

from data_loaderd import RescaleT
from data_loaderd import RandomCrop
from data_loaderd import RandomHorizontalFlip
from data_loaderd import ToTensorLab
from data_loaderd import SalObjDataset

from lr_scheduler import LR_Scheduler

from lovasz_losses import lovasz_hinge


import pytorch_ssim
import pytorch_iou

import torch.backends.cudnn as cudnn

from apex import amp
# from torch.cuda.amp import autocast as autocast, GradScaler
import torch.nn.functional as F

cudnn.enabled = True
cudnn.benchmark = True
torch.cuda.set_device(0)

# ------- 1. define loss function --------

bce_loss = nn.BCELoss(size_average=True)
ssim_loss = pytorch_ssim.SSIM(window_size=11, size_average=True)
iou_loss = pytorch_iou.IOU(size_average=True)
bce_loss_with_logits = nn.BCEWithLogitsLoss(size_average=True)
# mse = nn.MSELoss(reduce=False)


def bce_ssim_loss(pred, target):
    bce_out = bce_loss(pred, target)
    ssim_out = 1 - ssim_loss(pred, target)
    iou_out = iou_loss(pred, target)

    loss = bce_out + ssim_out + iou_out

    return loss

def bce_iou_loss(pred, target):
    pred  = torch.sigmoid(pred)

    bce_out = bce_loss(pred, target)
    # ssim_out = 1 - ssim_loss(pred, target)
    iou_out = iou_loss(pred, target)

    loss = bce_out + iou_out

    return loss

def structure_loss(pred, mask):
    weit  = 1+5*torch.abs(F.avg_pool2d(mask, kernel_size=31, stride=1, padding=15)-mask)
    wbce  = F.binary_cross_entropy_with_logits(pred, mask, reduce='none')
    wbce  = (weit*wbce).sum(dim=(2,3))/weit.sum(dim=(2,3))

    pred  = torch.sigmoid(pred)
    inter = ((pred*mask)*weit).sum(dim=(2,3))
    union = ((pred+mask)*weit).sum(dim=(2,3))
    wiou  = 1-(inter+1)/(union-inter+1)
    return (wbce+wiou).mean()

def muti_bce_loss_fusion( d1, d2, d3, d4, labels_v):
    # mse_loss = torch.sum(torch.mean(((ref - reflections_v)**2).view(is_ref_v.numel(), -1), 1) * is_ref_v)
    # loss0 = structure_loss(d0, labels_v)
    # loss1 = lovasz_hinge(d1.squeeze(), labels_v.squeeze())
    # loss2 = lovasz_hinge(d2.squeeze(), labels_v.squeeze())
    # loss3 = lovasz_hinge(d3.squeeze(), labels_v.squeeze())
    # loss4 = lovasz_hinge(d4.squeeze(), labels_v.squeeze())
    # print(torch.unique(labels_v))
    # print(torch.unique(d1))
    # print(torch.unique(d2))
    # print(torch.unique(d3))
    # print(torch.unique(d4))

    loss1 = bce_iou_loss(d1, labels_v)
    loss2 = bce_iou_loss(d2, labels_v)
    loss3 = bce_iou_loss(d3, labels_v)
    loss4 = bce_iou_loss(d4, labels_v)
    loss = loss1 + loss2 + loss3 + loss4
    print("l1: %3f, l2: %3f, l3: %3f, l4: %3f\n" % (
          loss1.item(), loss2.item(), loss3.item(), loss4.item()))
    return loss1, loss


def muti_bce_loss_fusion_withdepth( d1, d2, d3, d4, depth1, depth2, depth3, depth4, labels_v):
    loss1 = bce_iou_loss(d1, labels_v)
    loss2 = bce_iou_loss(d2, labels_v)
    loss3 = bce_iou_loss(d3, labels_v)
    loss4 = bce_iou_loss(d4, labels_v)

    depth_loss1 = bce_iou_loss(depth1, labels_v)
    depth_loss2 = bce_iou_loss(depth2, labels_v)
    depth_loss3 = bce_iou_loss(depth3, labels_v)
    depth_loss4 = bce_iou_loss(depth4, labels_v)

    loss = loss1 + loss2 + loss3 + loss4 + depth_loss1 + depth_loss2 + depth_loss3 + depth_loss4 
    print("l1: %3f, l2: %3f, l3: %3f, l4: %3f, depth_l1: %3f, depth_l2: %3f, depth_l3: %3f, depth_l4: %3f\n\n" % (loss1.item(), loss2.item(), loss3.item(), loss4.item(), \
        depth_loss1.item(), depth_loss2.item(), depth_loss3.item(), depth_loss4.item()))
    return loss1, loss


# ------- 2. set the directory of training dataset --------
# from sys import argv

# script, exp_name = argv

# exp_name = 'depth_20220107'
# print(exp_name)

# import importlib

# GlassNet = importlib.import_module(exp_name)
# GlassNet = getattr(importlib.import_module(exp_name, "GlassNet"), "GlassNet")
# my_class = getattr(module, 'MyClass')
# GlassNet = importlib.import_module("GlassNet", exp_name)

from model_depth import GlassNet

exp_name = 'model'


data_dir = '../datasets/glass_depth/train/'
tra_image_dir = 'images/'
tra_label_dir = 'masks/'
tra_depth_dir = 'depths/'

image_ext = '.jpg'
label_ext = '.png'

model_dir = "./saved_models/{}/".format(exp_name)

epoch_num = 200
batch_size_train = 14
batch_size_val = 1
train_num = 0
val_num = 0
lr = 1e-4
train_size = 384

tra_img_name_list = glob.glob(data_dir + tra_image_dir + '*' + image_ext)

tra_lbl_name_list = []
tra_depth_name_list = []
for img_path in tra_img_name_list:
    img_name = img_path.split("/")[-1]

    aaa = img_name.split(".")
    bbb = aaa[0:-1]
    imidx = bbb[0]
    for i in range(1, len(bbb)):
        imidx = imidx + "." + bbb[i]

    tra_lbl_name_list.append(data_dir + tra_label_dir + imidx + label_ext)
    tra_depth_name_list.append(data_dir + tra_depth_dir + imidx + label_ext)

print("---")
print("train images: ", len(tra_img_name_list))
assert len(tra_img_name_list) == len(tra_lbl_name_list)
print("---")

train_num = len(tra_img_name_list)

salobj_dataset = SalObjDataset(
    img_name_list=tra_img_name_list,
    lbl_name_list=tra_lbl_name_list,
    dep_name_list=tra_depth_name_list,
    transform=transforms.Compose([
        RescaleT(400),
        RandomCrop(384),
        RandomHorizontalFlip(0.5),
        ToTensorLab(flag=0)]))

salobj_dataloader = DataLoader(salobj_dataset, batch_size=batch_size_train, shuffle=True, num_workers=16)

# ------- 3. define model --------
# define the net
net = GlassNet('resnext_101_32x4d.pth')
print("Glassnet Defined.")
if torch.cuda.is_available():
    net.cuda()

# ------- 4. define optimizer --------
print("---define optimizer...")
optimizer = optim.Adam(net.parameters(), lr=lr)
# , betas=(0.9, 0.999), eps=1e-08, weight_decay=0)
# optimizer = optim.SGD(net.parameters(), lr=lr, momentum=0.9, weight_decay=5e-4)

# ------- 5. training process --------
print("---start training...")
ite_num = 0
running_loss = 0.0
running_tar_loss = 0.0
ite_num4val = 0

# poly_scheduler = LR_Scheduler('poly', base_lr=lr, num_epochs=epoch_num,
#                               iters_per_epoch=train_num // batch_size_train)

net.train()

net, optimizer = amp.initialize(net, optimizer, opt_level='O2')

def adjust_lr(optimizer, init_lr, epoch, decay_rate=0.1, decay_epoch=5):
    decay = decay_rate ** (epoch // decay_epoch)
    for param_group in optimizer.param_groups:
        param_group['lr'] *= decay


# scaler = GradScaler()
size_rates = [0.75,1,1.25]  # multi-scale training

for epoch in range(1, epoch_num+1):
    
    for i, data in enumerate(salobj_dataloader):

        for rate in size_rates:
            ite_num = ite_num + 1
            ite_num4val = ite_num4val + 1

            # poly_scheduler(optimizer, ite_num, epoch)

            inputs, labels, depth, depth_missing = data['image'], data['label'], data['depth'], data['depthmissing']

            inputs = inputs.type(torch.FloatTensor)
            labels = labels.type(torch.FloatTensor)
            depth = depth.type(torch.FloatTensor)
            depth_missing = depth_missing.type(torch.FloatTensor)

            # wrap them in Variable
            if torch.cuda.is_available():
                inputs_v, labels_v, depth_v, depth_missing_v = Variable(inputs.cuda(), requires_grad=False), \
                                                            Variable(labels.cuda(), requires_grad=False), \
                                                            Variable(depth.cuda(), requires_grad=False), \
                                                            Variable(depth_missing.cuda(), requires_grad=False)
            else:
                inputs_v, labels_v, depth_v, depth_missing_v = Variable(inputs, requires_grad=False), \
                                                            Variable(labels, requires_grad=False), \
                                                            Variable(depth, requires_grad=False), \
                                                            Variable(depth_missing, requires_grad=False)

            # y zero the parameter gradients

            trainsize = int(round(train_size * rate / 32) * 32)
            if rate != 1:
                inputs_v = F.upsample(inputs_v, size=(trainsize, trainsize), mode='bilinear',
                                          align_corners=True)
                labels_v = F.upsample(labels_v, size=(trainsize, trainsize), mode='bilinear', align_corners=True)
                depth_v = F.upsample(depth_v, size=(trainsize, trainsize), mode='bilinear', align_corners=True)
                depth_missing_v = F.upsample(depth_missing_v, size=(trainsize, trainsize), mode='bilinear', align_corners=True)


            optimizer.zero_grad()

            # forward + backward + optimize
            # with autocast():
            d1, d2, d3, d4, depth1, depth2, depth3, depth4 = net(inputs_v, depth_v, depth_missing_v)
            loss2, loss = muti_bce_loss_fusion_withdepth(d1, d2, d3, d4, depth1, depth2, depth3, depth4, labels_v)
            
            
            with amp.scale_loss(loss, optimizer) as scale_loss:
                scale_loss.backward()
            # scaler.scale(loss).backward()

            # loss.backward()
            optimizer.step()

            # # print statistics
            if rate == 1:
                running_loss += loss.item()
                running_tar_loss += loss2.item()

                # del temporary outputs and loss
                # del d1, d2, d3, d4, loss2, loss

                print('lr: %3f' % (optimizer.param_groups[0]['lr']))
                print("[epoch: %3d/%3d, batch: %5d/%5d, ite: %d] train loss: %3f, tar: %3f " % (
                    epoch, epoch_num, (i + 1) * batch_size_train, train_num, ite_num, running_loss / ite_num4val,
                    running_tar_loss / ite_num4val))

    
    adjust_lr(optimizer, lr, epoch, 0.9, int(epoch_num * 0.6))

    if epoch % 10 == 0:
        if not os.path.exists(model_dir):
            os.makedirs(model_dir)
        torch.save(net.state_dict(), model_dir + "basnet_bsi_itr_%d_train_%3f_tar_%3f.pth" % (
            epoch, running_loss / ite_num4val, running_tar_loss / ite_num4val))
        running_loss = 0.0
        running_tar_loss = 0.0
        net.train()  # resume train
        ite_num4val = 0
    


print('-------------Congratulations! Training Done!!!-------------')
